@@ -19,7 +19,6 @@ Support for running 'shell commands'
 
 from __future__ import absolute_import
 from __future__ import print_function
-from future.builtins import range
 from future.utils import PY3
 from future.utils import iteritems
 from future.utils import string_types
@@ -34,7 +33,6 @@ import subprocess
 import sys
 import traceback
 from codecs import getincrementaldecoder
-from collections import deque
 from tempfile import NamedTemporaryFile
 
 from twisted.internet import defer
@@ -88,7 +86,8 @@ def shell_quote(cmd_list, unicode_encoding='utf-8'):
     if runtime.platformType == 'win32':
         return win32_batch_quote(cmd_list, unicode_encoding)
 
-    import pipes  # only available on unix
+    # only available on unix
+    import pipes   # pylint: disable=import-outside-toplevel
 
     def quote(e):
         if not e:
@@ -101,10 +100,12 @@ def shell_quote(cmd_list, unicode_encoding='utf-8'):
 class LogFileWatcher(object):
     POLL_INTERVAL = 2
 
-    def __init__(self, command, name, logfile, follow=False):
+    def __init__(self, command, name, logfile, follow=False, poll=True):
         self.command = command
         self.name = name
         self.logfile = logfile
+        decoderFactory = getincrementaldecoder(self.command.unicode_encoding)
+        self.logDecode = decoderFactory(errors='replace')
 
         log.msg("LogFileWatcher created to watch {0}".format(logfile))
         # we are created before the ShellCommand starts. If the logfile we're
@@ -118,7 +119,7 @@ class LogFileWatcher(object):
         self.follow = follow
 
         # every 2 seconds we check on the file again
-        self.poller = task.LoopingCall(self.poll)
+        self.poller = task.LoopingCall(self.poll) if poll else None
 
     def start(self):
         self.poller.start(self.POLL_INTERVAL).addErrback(self._cleanupPoll)
@@ -158,12 +159,20 @@ class LogFileWatcher(object):
             if self.follow:
                 self.f.seek(s[2], 0)
             self.started = True
+
+        # Mac OS X and Linux differ in behaviour when reading from a file that has previously
+        # reached EOF. On Linux, any new data that has been appended to the file will be returned.
+        # On Mac OS X, the empty string will always be returned. Seeking to the current position
+        # in the file resets the EOF flag on Mac OS X and will allow future reads to work as
+        # intended.
         self.f.seek(self.f.tell(), 0)
+
         while True:
             data = self.f.read(10000)
             if not data:
                 return
-            self.command.addLogfile(self.name, data)
+            decodedData = self.logDecode.decode(data)
+            self.command.addLogfile(self.name, decodedData)
 
 
 if runtime.platformType == 'posix':
@@ -189,8 +198,7 @@ class RunProcessPP(protocol.ProcessProtocol):
         self.pending_stdin = b""
         self.stdin_finished = False
         self.killed = False
-        decoderFactory = getincrementaldecoder(
-            self.command.builder.unicode_encoding)
+        decoderFactory = getincrementaldecoder(self.command.unicode_encoding)
         self.stdoutDecode = decoderFactory(errors='replace')
         self.stderrDecode = decoderFactory(errors='replace')
 
@@ -258,15 +266,8 @@ class RunProcess(object):
     This is a helper class, used by worker commands to run programs in a child
     shell.
     """
-
     BACKUP_TIMEOUT = 5
     interruptSignal = "KILL"
-    CHUNK_LIMIT = 128 * 1024
-
-    # Don't send any data until at least BUFFER_SIZE bytes have been collected
-    # or BUFFER_TIMEOUT elapsed
-    BUFFER_SIZE = 64 * 1024
-    BUFFER_TIMEOUT = 5
 
     # For sending elapsed time:
     startTime = None
@@ -280,8 +281,7 @@ class RunProcess(object):
     # Then changes to the system clock during a run wouldn't effect the "elapsed
     # time" results.
 
-    def __init__(self, builder, command,
-                 workdir, environ=None,
+    def __init__(self, command, workdir, unicode_encoding, send_update, environ=None,
                  sendStdout=True, sendStderr=True, sendRC=True,
                  timeout=None, maxTime=None, sigtermTime=None,
                  initialStdin=None, keepStdout=False, keepStderr=False,
@@ -303,7 +303,6 @@ class RunProcess(object):
         if logfiles is None:
             logfiles = {}
 
-        self.builder = builder
         if isinstance(command, list):
             def obfus(w):
                 if (isinstance(w, tuple) and len(w) == 3 and
@@ -326,9 +325,9 @@ class RunProcess(object):
             if isinstance(cmd, (tuple, list)):
                 for i, a in enumerate(cmd):
                     if isinstance(a, text_type):
-                        cmd[i] = a.encode(self.builder.unicode_encoding)
+                        cmd[i] = a.encode(unicode_encoding)
             elif isinstance(cmd, text_type):
-                cmd = cmd.encode(self.builder.unicode_encoding)
+                cmd = cmd.encode(unicode_encoding)
             return cmd
 
         self.command = to_bytes(util.Obfuscated.get_real(command))
@@ -339,6 +338,8 @@ class RunProcess(object):
         self.sendRC = sendRC
         self.logfiles = logfiles
         self.workdir = workdir
+        self.unicode_encoding = unicode_encoding
+        self.send_update = send_update
         self.process = None
         if not os.path.exists(workdir):
             os.makedirs(workdir)
@@ -387,10 +388,6 @@ class RunProcess(object):
         self.keepStdout = keepStdout
         self.keepStderr = keepStderr
 
-        self.buffered = deque()
-        self.buflen = 0
-        self.sendBuffersTimer = None
-
         assert usePTY in (True, False), \
             "Unexpected usePTY argument value: {!r}. Expected boolean.".format(
                 usePTY)
@@ -402,8 +399,7 @@ class RunProcess(object):
         # and for .closeStdin to matter, we must use a pipe, not a PTY
         if runtime.platformType != "posix" or initialStdin is not None:
             if self.usePTY:
-                self.sendStatus(
-                    {'header': "WARNING: disabling usePTY for this command"})
+                self.send_update([('header', "WARNING: disabling usePTY for this command")])
             self.usePTY = False
 
         # use an explicit process group on POSIX, noting that usePTY always implies
@@ -433,9 +429,6 @@ class RunProcess(object):
     def __repr__(self):
         return "<{0} '{1}'>".format(self.__class__.__name__, self.fake_command)
 
-    def sendStatus(self, status):
-        self.builder.sendUpdate(status)
-
     def start(self):
         # return a Deferred which fires (with the exit code) when the command
         # completes
@@ -446,13 +439,13 @@ class RunProcess(object):
         self.deferred = defer.Deferred()
         try:
             self._startCommand()
-        except Exception:
+        except Exception as e:
             log.err(failure.Failure(), "error in RunProcess._startCommand")
-            self._addToBuffers('stderr', "error in RunProcess._startCommand\n")
-            self._addToBuffers('stderr', traceback.format_exc())
-            self._sendBuffers()
+            self.send_update([('stderr', "error in RunProcess._startCommand (%s)\n" % str(e))])
+
+            self.send_update([('stderr', traceback.format_exc())])
             # pretend it was a shell error
-            self.deferred.errback(AbandonChain(-1))
+            self.deferred.errback(AbandonChain(-1, 'Got exception (%s)' % str(e)))
         return self.deferred
 
     def _startCommand(self):
@@ -464,7 +457,7 @@ class RunProcess(object):
         self.pp = RunProcessPP(self)
 
         self.using_comspec = False
-        self.command = unicode2bytes(self.command, encoding=self.builder.unicode_encoding)
+        self.command = unicode2bytes(self.command, encoding=self.unicode_encoding)
         if isinstance(self.command, bytes):
             if runtime.platformType == 'win32':
                 # allow %COMSPEC% to have args
@@ -486,7 +479,7 @@ class RunProcess(object):
             # handle path searching, etc.
             if (runtime.platformType == 'win32' and
                 not (bytes2unicode(self.command[0],
-                     self.builder.unicode_encoding).lower().endswith(".exe") and
+                     self.unicode_encoding).lower().endswith(".exe") and
                      os.path.isabs(self.command[0]))):
                 # allow %COMSPEC% to have args
                 argv = os.environ['COMSPEC'].split()
@@ -498,9 +491,9 @@ class RunProcess(object):
                 argv = self.command
             # Attempt to format this for use by a shell, although the process
             # isn't perfect
-            display = shell_quote(self.fake_command, self.builder.unicode_encoding)
+            display = shell_quote(self.fake_command, self.unicode_encoding)
 
-        display = bytes2unicode(display, self.builder.unicode_encoding)
+        display = bytes2unicode(display, self.unicode_encoding)
 
         # $PWD usually indicates the current directory; spawnProcess may not
         # update this value, though, so we set it explicitly here.  This causes
@@ -511,7 +504,7 @@ class RunProcess(object):
         # self.stdin is handled in RunProcessPP.connectionMade
 
         log.msg(u" " + display)
-        self._addToBuffers(u'header', display + u"\n")
+        self.send_update([(u'header', display + u"\n")])
 
         # then comes the secondary information
         msg = u" in dir {0}".format(self.workdir)
@@ -528,35 +521,37 @@ class RunProcess(object):
                 unit = u"secs"
             msg += u" (maxTime {0} {1})".format(self.maxTime, unit)
         log.msg(u" " + msg)
-        self._addToBuffers(u'header', msg + u"\n")
+        self.send_update([(u'header', msg + u"\n")])
 
         msg = " watching logfiles {0}".format(self.logfiles)
         log.msg(" " + msg)
-        self._addToBuffers('header', msg + u"\n")
+        self.send_update([('header', msg + u"\n")])
 
         # then the obfuscated command array for resolving unambiguity
         msg = u" argv: {0}".format(self.fake_command)
         log.msg(u" " + msg)
-        self._addToBuffers('header', msg + u"\n")
+        self.send_update([('header', msg + u"\n")])
 
         # then the environment, since it sometimes causes problems
         if self.logEnviron:
             msg = u" environment:\n"
             env_names = sorted(self.environ.keys())
             for name in env_names:
-                msg += u"  {0}={1}\n".format(bytes2unicode(name, encoding=self.builder.unicode_encoding),
-                                             bytes2unicode(self.environ[name], encoding=self.builder.unicode_encoding))
+                msg += u"  {0}={1}\n".format(bytes2unicode(name,
+                                                           encoding=self.unicode_encoding),
+                                             bytes2unicode(self.environ[name],
+                                                           encoding=self.unicode_encoding))
             log.msg(u" environment:\n{0}".format(pprint.pformat(self.environ)))
-            self._addToBuffers(u'header', msg)
+            self.send_update([(u'header', msg)])
 
         if self.initialStdin:
             msg = u" writing {0} bytes to stdin".format(len(self.initialStdin))
             log.msg(u" " + msg)
-            self._addToBuffers(u'header', msg + u"\n")
+            self.send_update([(u'header', msg + u"\n")])
 
         msg = u" using PTY: {0}".format(bool(self.usePTY))
         log.msg(u" " + msg)
-        self._addToBuffers(u'header', msg + u"\n")
+        self.send_update([(u'header', msg + u"\n")])
 
         # put data into stdin and close it, if necessary.  This will be
         # buffered until connectionMade is called
@@ -616,7 +611,7 @@ class RunProcess(object):
         # In PY3, it needs str which is unicode and its encoding can be specified.
         if PY3:
             tf = NamedTemporaryFile(mode='w+', dir='.', suffix=".bat",
-                                    delete=False, encoding=self.builder.unicode_encoding)
+                                    delete=False, encoding=self.unicode_encoding)
         else:
             tf = NamedTemporaryFile(mode='w+', dir='.', suffix=".bat",
                                     delete=False)
@@ -624,9 +619,9 @@ class RunProcess(object):
         # echo off hides this cheat from the log files.
         tf.write(u"@echo off\n")
         if isinstance(self.command, (string_types, bytes)):
-            tf.write(bytes2NativeString(self.command, self.builder.unicode_encoding))
+            tf.write(bytes2NativeString(self.command, self.unicode_encoding))
         else:
-            tf.write(win32_batch_quote(self.command, self.builder.unicode_encoding))
+            tf.write(win32_batch_quote(self.command, self.unicode_encoding))
         tf.close()
 
         argv = os.environ['COMSPEC'].split()  # allow %COMSPEC% to have args
@@ -642,118 +637,9 @@ class RunProcess(object):
         return reactor.spawnProcess(processProtocol, executable, argv, env,
                                     path, usePTY=usePTY)
 
-    def _chunkForSend(self, data):
-        """
-        limit the chunks that we send over PB to 128k, since it has a hardwired
-        string-size limit of 640k.
-        """
-        LIMIT = self.CHUNK_LIMIT
-        for i in range(0, len(data), LIMIT):
-            yield data[i:i + LIMIT]
-
-    def _collapseMsg(self, msg):
-        """
-        Take msg, which is a dictionary of lists of output chunks, and
-        concatenate all the chunks into a single string
-        """
-        retval = {}
-        for logname in msg:
-            data = u""
-            for m in msg[logname]:
-                m = bytes2unicode(m, self.builder.unicode_encoding)
-                data += m
-            if isinstance(logname, tuple) and logname[0] == 'log':
-                retval['log'] = (logname[1], data)
-            else:
-                retval[logname] = data
-        return retval
-
-    def _sendMessage(self, msg):
-        """
-        Collapse and send msg to the master
-        """
-        if not msg:
-            return
-        msg = self._collapseMsg(msg)
-        self.sendStatus(msg)
-
-    def _bufferTimeout(self):
-        self.sendBuffersTimer = None
-        self._sendBuffers()
-
-    def _sendBuffers(self):
-        """
-        Send all the content in our buffers.
-        """
-        msg = {}
-        msg_size = 0
-        lastlog = None
-        logdata = []
-        while self.buffered:
-            # Grab the next bits from the buffer
-            logname, data = self.buffered.popleft()
-
-            # If this log is different than the last one, then we have to send
-            # out the message so far.  This is because the message is
-            # transferred as a dictionary, which makes the ordering of keys
-            # unspecified, and makes it impossible to interleave data from
-            # different logs.  A future enhancement could be to change the
-            # master to support a list of (logname, data) tuples instead of a
-            # dictionary.
-            # On our first pass through this loop lastlog is None
-            if lastlog is None:
-                lastlog = logname
-            elif logname != lastlog:
-                self._sendMessage(msg)
-                msg = {}
-                msg_size = 0
-            lastlog = logname
-
-            logdata = msg.setdefault(logname, [])
-
-            # Chunkify the log data to make sure we're not sending more than
-            # CHUNK_LIMIT at a time
-            for chunk in self._chunkForSend(data):
-                if not chunk:
-                    continue
-                logdata.append(chunk)
-                msg_size += len(chunk)
-                if msg_size >= self.CHUNK_LIMIT:
-                    # We've gone beyond the chunk limit, so send out our
-                    # message.  At worst this results in a message slightly
-                    # larger than (2*CHUNK_LIMIT)-1
-                    self._sendMessage(msg)
-                    msg = {}
-                    logdata = msg.setdefault(logname, [])
-                    msg_size = 0
-        self.buflen = 0
-        if logdata:
-            self._sendMessage(msg)
-        if self.sendBuffersTimer:
-            if self.sendBuffersTimer.active():
-                self.sendBuffersTimer.cancel()
-            self.sendBuffersTimer = None
-
-    def _addToBuffers(self, logname, data):
-        """
-        Add data to the buffer for logname
-        Start a timer to send the buffers if BUFFER_TIMEOUT elapses.
-        If adding data causes the buffer size to grow beyond BUFFER_SIZE, then
-        the buffers will be sent.
-        """
-        n = len(data)
-
-        self.buflen += n
-        self.buffered.append((logname, data))
-        if self.buflen > self.BUFFER_SIZE:
-            self._sendBuffers()
-        elif not self.sendBuffersTimer:
-            self.sendBuffersTimer = self._reactor.callLater(
-                self.BUFFER_TIMEOUT, self._bufferTimeout)
-
     def addStdout(self, data):
         if self.sendStdout:
-            self._addToBuffers('stdout', data)
+            self.send_update([('stdout', data)])
 
         if self.keepStdout:
             self.stdout += data
@@ -762,7 +648,7 @@ class RunProcess(object):
 
     def addStderr(self, data):
         if self.sendStderr:
-            self._addToBuffers('stderr', data)
+            self.send_update([('stderr', data)])
 
         if self.keepStderr:
             self.stderr += data
@@ -770,7 +656,7 @@ class RunProcess(object):
             self.ioTimeoutTimer.reset(self.timeout)
 
     def addLogfile(self, name, data):
-        self._addToBuffers(('log', name), data)
+        self.send_update([('log', (name, data))])
 
         if self.ioTimeoutTimer:
             self.ioTimeoutTimer.reset(self.timeout)
@@ -782,15 +668,13 @@ class RunProcess(object):
         for w in self.logFileWatchers:
             # this will send the final updates
             w.stop()
-        self._sendBuffers()
         if sig is not None:
             rc = -1
         if self.sendRC:
             if sig is not None:
-                self.sendStatus(
-                    {'header': "process killed by signal {0}\n".format(sig)})
-            self.sendStatus({'rc': rc})
-        self.sendStatus({'header': "elapsedTime={0:0.6f}\n".format(self.elapsedTime)})
+                self.send_update([('header', "process killed by signal {0}\n".format(sig))])
+            self.send_update([('rc', rc)])
+        self.send_update([('header', "elapsedTime={0:0.6f}\n".format(self.elapsedTime))])
         self._cancelTimers()
         d = self.deferred
         self.deferred = None
@@ -800,7 +684,6 @@ class RunProcess(object):
             log.msg("Hey, command {0} finished twice".format(self))
 
     def failed(self, why):
-        self._sendBuffers()
         log.msg("RunProcess.failed: command failed: {0}".format(why))
         self._cancelTimers()
         d = self.deferred
@@ -888,18 +771,10 @@ class RunProcess(object):
                 log.msg("interruptSignal==None, only pretending to kill child")
             elif self.process.pid is not None:
                 if interruptSignal == "TERM":
-                    log.msg("using TASKKILL PID /T to kill pid {0}".format(
-                            self.process.pid))
-                    subprocess.check_call(
-                        "TASKKILL /PID {0} /T".format(self.process.pid))
-                    log.msg("taskkill'd pid {0}".format(self.process.pid))
+                    self._taskkill(self.process.pid, force=False)
                     hit = 1
                 elif interruptSignal == "KILL":
-                    log.msg("using TASKKILL PID /F /T to kill pid {0}".format(
-                            self.process.pid))
-                    subprocess.check_call(
-                        "TASKKILL /F /PID {0} /T".format(self.process.pid))
-                    log.msg("taskkill'd pid {0}".format(self.process.pid))
+                    self._taskkill(self.process.pid, force=True)
                     hit = 1
 
         # try signalling the process itself (works on Windows too, sorta)
@@ -920,14 +795,33 @@ class RunProcess(object):
 
         return hit
 
+    def _taskkill(self, pid, force):
+        try:
+            if force:
+                cmd = "TASKKILL /F /PID {0} /T".format(pid)
+            else:
+                cmd = "TASKKILL /PID {0} /T".format(pid)
+
+            log.msg("using {0} to kill pid {1}".format(cmd, pid))
+            subprocess.check_call(cmd)
+            log.msg("taskkill'd pid {0}".format(pid))
+
+        except subprocess.CalledProcessError as e:
+            # taskkill may return 128 as exit code when the child has already exited. We can't
+            # handle this race condition in any other way than just interpreting the kill action
+            # as successful
+            if e.returncode == 128:
+                log.msg("taskkill didn't find pid {0} to kill".format(pid))
+            else:
+                log.msg("taskkill failed to kill process {0}: {1}".format(pid, e))
+
     def kill(self, msg):
         # This may be called by the timeout, or when the user has decided to
         # abort this build.
-        self._sendBuffers()
         self._cancelTimers()
         msg += ", attempting to kill"
         log.msg(msg)
-        self.sendStatus({'header': "\n" + msg + "\n"})
+        self.send_update([('header', "\n" + msg + "\n")])
 
         # let the PP know that we are killing it, so that it can ensure that
         # the exit status comes out right
@@ -947,15 +841,13 @@ class RunProcess(object):
                 " finish anyway")
         self.killTimer = None
         signalName = "SIG" + self.interruptSignal
-        self.sendStatus({'header': signalName + " failed to kill process\n"})
+        self.send_update([('header', signalName + " failed to kill process\n")])
         if self.sendRC:
-            self.sendStatus({'header': "using fake rc=-1\n"})
-            self.sendStatus({'rc': -1})
+            self.send_update([('header', "using fake rc=-1\n"), ('rc', -1)])
         self.failed(RuntimeError(signalName + " failed to kill process"))
 
     def _cancelTimers(self):
-        for timerName in ('ioTimeoutTimer', 'killTimer', 'maxTimeoutTimer',
-                          'sendBuffersTimer', 'sigtermTimer'):
+        for timerName in ('ioTimeoutTimer', 'killTimer', 'maxTimeoutTimer', 'sigtermTimer'):
             timer = getattr(self, timerName, None)
             if timer:
                 timer.cancel()
